/*
 * Written by Sebastian Tilders
 * Licensed under the GPL (General Public License)
 */
#include "../header/typedefs.h"
#include "../header/display.h"
#include "../header/interrupts.h"
#include "../header/ports.h"

#define GDT_ENTRIES 6
#define IDT_ENTRIES 254
#define IDT_TASK_GATE 0x5
#define IDT_INT_GATE 0x6
#define IDT_TRAP_GATE 0x7
struct task_state_segment
{
  uint32_t previous_task_link;
  uint32_t esp0;
  uint32_t ss0;
  uint32_t esp1;
  uint32_t ss1;
  uint32_t esp2;
  uint32_t ss2;
  uint32_t cr3;
  uint32_t eip;
  uint32_t eflags;
  uint32_t eax;
  uint32_t ecx;
  uint32_t edx;
  uint32_t ebx;
  uint32_t esp;
  uint32_t ebp;
  uint32_t esi;
  uint32_t edi;
  uint32_t es;
  uint32_t cs;
  uint32_t ss;
  uint32_t ds;
  uint32_t fs;
  uint32_t gs;
  uint32_t ldt_segment_selector;
  uint32_t io_map_base_adress;
};
struct task_state_segment tss = {
  .previous_task_link = 0,
  .esp0 = 0,
  .ss0 = 0x10,
  .esp1 = 0,
  .ss1 = 0x10,
  .esp2 = 0,
  .ss2 = 0x10,
  .io_map_base_adress = 0x0,
};
static uint64_t gdt[GDT_ENTRIES];
// IDT
static uint64_t idt[IDT_ENTRIES];
// Selektoren laden
extern void load_selectors(void);
// GDT Pointer
struct {
    uint16_t length;
    void* pointer;
} __attribute__((packed)) gdtp = {
    .length = GDT_ENTRIES * 8,
    .pointer = gdt,
};

// IDT Pointer
struct {
	uint16_t length;
	void* pointer;
} __attribute__((packed)) idtp = {
	.length = (IDT_ENTRIES * 8) - 1,
	.pointer = idt,
};

void null_handler(void)
{
print("ENDE!", DSP_COLOR_RED);
while(1)
{
asm volatile("cli;hlt");
}
}

// GDT Eintrag setzen
void set_entry_gdt(int i, unsigned int base, unsigned int limit, int flags)
{
    gdt[i] = limit & 0xffffLL;
    gdt[i] |= (base & 0xffffffLL) << 16;
    gdt[i] |= (flags & 0xffLL) << 40;
    gdt[i] |= ((limit >> 16) & 0xfLL) << 48;
    gdt[i] |= ((flags >> 8 )& 0xffLL) << 52;
    gdt[i] |= ((base >> 24) & 0xffLL) << 56;
}
// IDT Eintrag setzen
void set_entry_idt(uint16_t i, uint16_t base, long* Offset, uint8_t ACS)
{
	idt[i] = (long) Offset & 0xffffLL;
	idt[i] |= base << 16;
	idt[i] |= (ACS & 0xffLL) << 40;
	idt[i] |= (((long)Offset >> 16) & 0xffffLL) << 48;
}

// GDT laden
void load_gdt(void)
{
	asm volatile("lgdt %0" : : "m" (gdtp));
	load_selectors();
	asm volatile("ltr %%ax" : : "a" (5 << 3));
}
// IDT laden
void load_idt(void)
{
	asm volatile("lidt %0" : : "m" (idtp));
}

/* Hier finden Sie Hilfsfunktionen */
	
// IDT Type + Rechte
uint8_t createIDT_ACS(uint8_t Type, uint8_t ACS)
{
	uint8_t Buffer_ACS = 0x0;
	Buffer_ACS = (Type & 0x7);
	Buffer_ACS |= (ACS & 0x1F) << 3;
	return Buffer_ACS;
}

void setup_idt(void)
{
int i;
for(i=0; i < IDT_ENTRIES;i++)
{
set_entry_idt(i,0x8,(long*)&null_handler,createIDT_ACS(IDT_INT_GATE, 0x11));
}
//IDT

//Exceptions
set_entry_idt(0,0x8, (long*)&intr_stub_0,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(1,0x8, (long*)&intr_stub_1,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(2,0x8, (long*)&intr_stub_2,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(3,0x8, (long*)&intr_stub_3,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(4,0x8, (long*)&intr_stub_4,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(5,0x8, (long*)&intr_stub_5,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(6,0x8, (long*)&intr_stub_6,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(7,0x8, (long*)&intr_stub_7,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(8,0x8, (long*)&intr_stub_error_code_8,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(9,0x8, (long*)&intr_stub_9,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(10,0x8, (long*)&intr_stub_error_code_10,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(11,0x8, (long*)&intr_stub_error_code_11,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(12,0x8, (long*)&intr_stub_error_code_12,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(13,0x8, (long*)&intr_stub_error_code_13,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(14,0x8, (long*)&intr_stub_error_code_14,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(15,0x8, (long*)&intr_stub_15,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(16,0x8, (long*)&intr_stub_16,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(17,0x8, (long*)&intr_stub_error_code_17,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(18,0x8, (long*)&intr_stub_18,createIDT_ACS(IDT_INT_GATE, 0x11));

// IRQs
set_entry_idt(32,0x8, (long*)&intr_stub_32,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(33,0x8, (long*)&intr_stub_33,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(34,0x8, (long*)&intr_stub_34,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(35,0x8, (long*)&intr_stub_35,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(36,0x8, (long*)&intr_stub_36,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(37,0x8, (long*)&intr_stub_37,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(38,0x8, (long*)&intr_stub_38,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(39,0x8, (long*)&intr_stub_39,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(40,0x8, (long*)&intr_stub_40,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(41,0x8, (long*)&intr_stub_41,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(42,0x8, (long*)&intr_stub_42,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(43,0x8, (long*)&intr_stub_43,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(44,0x8, (long*)&intr_stub_44,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(45,0x8, (long*)&intr_stub_45,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(46,0x8, (long*)&intr_stub_46,createIDT_ACS(IDT_INT_GATE, 0x11));
set_entry_idt(47,0x8, (long*)&intr_stub_47,createIDT_ACS(IDT_INT_GATE, 0x11));
// Syscall
set_entry_idt(48,0x8, (long*)&intr_stub_48,createIDT_ACS(IDT_TRAP_GATE, 0x11));
load_idt();
return;
}
void setup_gdt(void)
{
set_entry_gdt(0,0,0,0);
set_entry_gdt(1,0x0,0xFFFFF,(0xC << 8) | 0x9A);
set_entry_gdt(2,0x0,0xFFFFF,(0xC << 8) | 0x92);
set_entry_gdt(3,0x0,0xFFFFF,(0xC << 8) | 0xFA);
set_entry_gdt(4,0x0,0xFFFFF,(0xC << 8) | 0xF2);
set_entry_gdt(5,(uint32_t) &tss,sizeof(tss),(0xC << 8) | 0x89);
load_gdt();
return;
}

void setup_pic(void)
{
// Master	@Dataport: 0x21; Befehlsport: 0x20
outb(0x20,0x11);
outb(0x21,0x20);	// IRQ start bei IDT-ISR 32
outb(0x21,0x04);
outb(0x21,0x01);

//Slave @Dataport: 0xA1; Befehlsport: 0xA0
outb(0xA0,0x11);
outb(0xA1,0x28);	// IRQ start bei IDT-ISR 40
outb(0xA1,0x02);
outb(0xA1,0x01);

// IRQs demaskieren
outb(0x20,0x0);			// Alle IRQs zulassen auf Master
outb(0xa0,0x0);			// Alle IRQs zulassen auf Slave
asm volatile("sti");
return;
}