/*
 * Written by Sebastian Tilders
 * Licensed under the GPL (General PUblic License).
 * Project Name: InfoOS
 */
#include "../header/typedefs.h"
#include "../header/interrupts.h"
#include "../header/string.h"
#include "../header/task.h"
#define TASK_CURRENT tasks.current_task
#define TASK_NUM_TASKS tasks.num_tasks
#define TASK_STACK tasks.task_status[TASK_NUM_TASKS].stack
#define TASK_USER_STACK tasks.task_status[TASK_NUM_TASKS].user_stack
#define TASK_STATE tasks.task_status[TASK_NUM_TASKS].task_state
struct task_list tasks;

// Interrupts, Multitasking
/* 
	Einen Task in den Anfangszustand versetzen, sprich alle register = 0 und Stack initalisieren
*/
struct cpu_state* init_ring0_task(uint8_t* stack, uint8_t* entry)
{
	struct cpu_state new_state = {
	.eax = 0,
	.ebx = 0,
	.ecx = 0,
	.edx = 0,
	.esi = 0,
	.edi = 0,
	.ebp = 0,
	.eip = (uint32_t) entry,
	.cs = 0x8,
	.eflags = 0x202,
	};
	struct cpu_state* state = (void*) (stack + 4096 - sizeof(new_state));
	*state = new_state;
	return state;
}
struct cpu_state* init_ring3_task(uint8_t *stack, uint8_t *userstack,  uint8_t *entry)
{
  struct cpu_state new_state = {
    .eax = 0,
    .ebx = 0,
    .ecx = 0,
    .edx = 0,
    .esi = 0,
    .edi = 0,
    .ebp = 0,
    .eip = (uint32_t) entry,
    .cs = 0x18 | 0x03,
    .ss = 0x20 | 0x03,
    .esp = (uint32_t) userstack + 4096,
    .eflags = 0x202,
  };
  struct cpu_state* state = (void*) (stack + 4096 - sizeof(new_state));
  *state = new_state;
  return state;
}
// Taskliste  @TODO: überarbeiten (PID etc. hinzufügen)
// ENDE

void init_multi(void)
{
	int i;
	tasks.current_task = -1;
	tasks.num_tasks = 0;
	for(i=0;i < 64;i++)
	{
		tasks.task_status[i].pid = 0;
	}
	return;
}
uint16_t create_task(uint8_t* process, uint16_t ring)
{
	tasks.task_status[TASK_NUM_TASKS].pid = TASK_NUM_TASKS;
	if(ring != 0)
	{
	  TASK_STATE = init_ring3_task(TASK_STACK,TASK_USER_STACK,process);
	}
	else
	{
	  TASK_STATE = init_ring0_task(TASK_STACK, process);
	}
	tasks.num_tasks++;
	return tasks.task_status[TASK_NUM_TASKS-1].pid;
}
// Taskliste abarbeiten @TODO: Priorit�ten einbauen.
struct cpu_state* schedule(struct cpu_state* cpu)
{
	if(tasks.current_task >= 0)
	{
	tasks.task_status[tasks.current_task].task_state = cpu;
	}
	tasks.current_task++;
	tasks.current_task %= tasks.num_tasks;
	cpu = tasks.task_status[tasks.current_task].task_state;
	return cpu;
}